"""Library json module."""

import json
import os
import re
import shutil
import subprocess
import tempfile
import warnings
from dataclasses import dataclass

from filelock import FileLock
from static_ffmpeg import add_paths


def clean_filename(filename: str) -> str:
    """
    Cleans a string to make it a valid directory name by removing emojis,
    special characters, and other non-ASCII characters, in addition to the
    previously specified invalid filename characters, while preserving the file extension.

    Args:
    - filename (str): The filename to be cleaned.

    Returns:
    - str: A cleaned-up string suitable for use as a filename.
    """
    # Split the filename into name and extension
    name_part, _, extension = filename.rpartition(".")

    # Remove emojis and special characters by allowing only a specific set of characters
    # This regex keeps letters, numbers, spaces, underscores, and hyphens.
    # You can adjust the regex as needed to include any additional characters.
    cleaned_name = re.sub(r"[^\w\s\-_]", "", name_part)

    # Replace spaces or consecutive dashes with a single underscore
    cleaned_name = re.sub(r"\s+|-+", "_", cleaned_name)

    # Remove leading or trailing whitespace (after replacing spaces with underscores, this might be redundant)
    cleaned_name = cleaned_name.strip()

    # Optional: Convert to lowercase to avoid issues with case-sensitive file systems
    # cleaned_name = cleaned_name.lower()

    # Optional: Trim the title to a maximum length (e.g., 255 characters)
    max_length = 255
    if len(cleaned_name) > max_length:
        cleaned_name = cleaned_name[:max_length]

    # Reattach the extension only if it was present
    if extension:
        return f"{cleaned_name}.{extension}"
    return cleaned_name


@dataclass
class VidEntry:
    """Video entry."""

    url: str
    title: str
    file_path: str

    def __init__(self, url: str, title: str, file_path: str | None = None) -> None:
        self.url = url
        self.title = title
        self.file_path = file_path or clean_filename(f"{title}.mp3")

    # needed for set membership
    def __hash__(self):
        return hash(self.url)

    def __eq__(self, other):
        return self.url == other.url

    def __repr__(self) -> str:
        data = self.to_dict()
        return json.dumps(data)

    def to_dict(self) -> dict:
        """Convert to dictionary."""
        return {
            "url": self.url,
            "title": self.title,
            "file_path": self.file_path,
        }

    @classmethod
    def from_dict(cls, data: dict) -> "VidEntry":
        """Create from dictionary."""
        filepath = data.get("file_path")
        assert filepath is not None
        filepath = clean_filename(filepath)
        return cls(url=data["url"], title=data["title"], file_path=filepath)

    @classmethod
    def serialize(cls, data: list["VidEntry"]) -> str:
        """Serialize to string."""
        json_data = [vid.to_dict() for vid in data]
        return json.dumps(json_data, indent=2)

    @classmethod
    def deserialize(cls, data: str) -> list["VidEntry"]:
        """Deserialize from string."""
        return [cls.from_dict(vid) for vid in json.loads(data)]


def find_missing_downloads(library_json_path: str) -> list[VidEntry]:
    """Find missing downloads."""
    pardir = os.path.dirname(library_json_path)
    out: list[VidEntry] = []
    lock = library_json_path + ".lock"
    with FileLock(lock):
        data = load_json(library_json_path)
        for vid in data:
            file_path = vid.file_path
            if not os.path.exists(os.path.join(pardir, file_path)):
                out.append(vid)
    return out


def load_json(file_path: str) -> list[VidEntry]:
    """Load json from file."""
    with open(file_path, encoding="utf-8", mode="r") as filed:
        data = filed.read()
    return VidEntry.deserialize(data)


def save_json(file_path: str, data: list[VidEntry]) -> None:
    """Save json to file."""
    json_out = VidEntry.serialize(data)
    with open(file_path, encoding="utf-8", mode="w") as filed:
        filed.write(json_out)


def yt_dlp_download_mp3(url: str, outmp3: str) -> None:
    """Download the youtube video as an mp3."""
    add_paths()
    par_dir = os.path.dirname(outmp3)
    if par_dir:
        os.makedirs(par_dir, exist_ok=True)

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_file = os.path.join(temp_dir, "temp.mp3")
        for _ in range(3):
            try:
                cmd_list: list[str] = [
                    "yt-dlp",
                    url,
                    "--extract-audio",
                    "--audio-format",
                    "mp3",
                    "--output",
                    temp_file,
                ]
                subprocess.run(cmd_list, check=True)
                shutil.copy(temp_file, outmp3)
                return
            except subprocess.CalledProcessError as cpe:
                print(f"Failed to download {url} as mp3: {cpe}")
                continue
        warnings.warn(f"Failed all attempts to download {url} as mp3.")


def merge_into_library(library_json_path: str, vids: list[VidEntry]) -> None:
    """Merge the vids into the library."""
    found_entries: list[VidEntry] = []
    for vid in vids:
        title = vid.title
        file_path = vid.file_path
        found_entries.append(VidEntry(url=vid.url, title=title, file_path=file_path))
    file_lock = library_json_path + ".lock"
    with FileLock(file_lock):
        existing_entries = load_json(library_json_path)
        for found in found_entries:
            if found not in existing_entries:
                existing_entries.append(found)
        save_json(library_json_path, existing_entries)


class Library:
    """Represents the library"""

    def __init__(self, library_json_path: str) -> None:
        self.library_json_path = library_json_path
        self.base_dir = os.path.dirname(library_json_path)
        pardir = os.path.dirname(library_json_path)
        if not os.path.exists(pardir):
            os.makedirs(pardir)
        if not os.path.exists(library_json_path):
            with open(library_json_path, encoding="utf-8", mode="w") as filed:
                filed.write("[]")

    def find_missing_downloads(self) -> list[VidEntry]:
        """Find missing downloads."""
        return find_missing_downloads(self.library_json_path)

    def load(self) -> list[VidEntry]:
        """Load json from file."""
        return load_json(self.library_json_path)

    def save(self, data: list[VidEntry]) -> None:
        """Save json to file."""
        save_json(self.library_json_path, data)

    def merge(self, vids: list[VidEntry]) -> None:
        """Merge the vids into the library."""
        merge_into_library(self.library_json_path, vids)

    def download_missing(self, download_limit: int = -1) -> None:
        """Download the missing files."""
        download_count = 0
        while True:
            if download_limit != -1 and download_count >= download_limit:
                break
            missing_downloads = self.find_missing_downloads()
            # make full paths
            if not missing_downloads:
                break
            vid = missing_downloads[0]
            next_url = vid.url
            next_mp3_path = os.path.join(self.base_dir, vid.file_path)
            print(f"\n#######################\n# Downloading missing file {next_url}: {next_mp3_path}\n" "###################")
            yt_dlp_download_mp3(url=next_url, outmp3=next_mp3_path)
            download_count += 1



"""
Fetcher for html
"""

import subprocess

# pylint: disable=line-too-long,missing-function-docstring,consider-using-f-string,too-many-locals,invalid-name,no-else-return
from typing import Optional

import requests

try:
    subprocess.check_output("curl --version", shell=True)
    USE_CURL = True
except BaseException:  # pylint: disable=broad-except
    USE_CURL = False


def fetch_html_using_request_lib(
    url: str,
    timeout: Optional[int] = None,
    user_agent: Optional[str] = None,
) -> str:
    timeout = timeout or 10
    # Workaround for long request time on windows
    # see https://github.com/psf/requests/issues/4023
    headers = {"Connection": "close"}
    # headers = {"Connection": "close"}
    if user_agent:
        headers["User-Agent"] = user_agent
    resp = requests.get(url, timeout=timeout, params={}, headers=headers)
    resp.raise_for_status()
    return resp.content.decode(resp.apparent_encoding)


def fetch_html_using_curl(url: str, timeout: Optional[int] = None) -> str:
    """Uses the curl library to do a fetch"""
    timeout = int(timeout or 10)
    out: bytes = subprocess.check_output(f"curl --max-time {timeout} -s -X GET {url}", shell=True)
    return out.decode("utf-8")


def fetch_html(url: str) -> str:
    if USE_CURL:
        return fetch_html_using_curl(url)
    else:
        return fetch_html_using_request_lib(url)



# pylint: disable=all

import unittest
from typing import List

from vidcrawler.date import now_local
from vidcrawler.video_info import VideoInfo


class VideoInfoTester(unittest.TestCase):
    def test_one(self) -> None:
        now_str = str(now_local().isoformat())
        vid_list = []
        vi = VideoInfo()
        vi.title = "test_channel"
        vi.img_src = "https://example.com/img_src.png"
        vi.description = "My video description"
        vi.iframe_src = "https://example/123/player.html"
        vi.duration = "1:03:53"
        vi.views = "1000"
        vi.date_published = "2021-12-21T03:30:19+00:00"
        vi.channel_url = "https://rumble.com/c/PeteSantilliInterviews"
        vi.channel_name = "Pete Santilli Interviews"
        vi.date_published = now_str
        vi.date_discovered = now_str
        vi.date_lastupdated = now_str
        vid_list.append(vi)
        data: List[List] = VideoInfo.to_compact_csv(vid_list)
        vid_list2: List[VideoInfo] = VideoInfo.from_compact_csv(data)
        vi2: VideoInfo = vid_list2[0]
        self.assertEqual(vi, vi2)


if __name__ == "__main__":
    unittest.main()



"""
    Scraper for gabtv.com
"""

# pylint: disable=line-too-long,missing-function-docstring,consider-using-f-string,too-many-locals,invalid-name,no-else-return,redefined-builtin,W0511

import json
import re
import subprocess
import sys
from typing import Dict, List

from bs4 import BeautifulSoup  # type: ignore

from .date import iso_fmt, now_local
from .error import log_error
from .video_info import VideoInfo

_PATTERN_DATA_EPISODE_ID = re.compile('data-episode-id="([^"]*)"')


def _fetch_html_using_curl(url: str) -> str:
    out: bytes = subprocess.check_output("curl --max-time 10 -s -X GET " + url, shell=True)
    return out.decode("utf-8")


def fetch_views(channel: str) -> Dict[str, str]:
    out: Dict[str, str] = {}
    html_url: str = f"https://tv.gab.com/channel/{channel}"
    html_doc: str = _fetch_html_using_curl(html_url)
    soup = BeautifulSoup(html_doc, "html.parser")
    top_dom = soup.find("div", {"class": "tv-channel-episode-list"})
    dom_episodes = top_dom.findAll("div")
    for i, dom_episode in enumerate(dom_episodes):
        str_episode = str(dom_episode)
        if "data-episode-id" not in str_episode:
            continue
        try:
            id = _PATTERN_DATA_EPISODE_ID.findall(str_episode)[0]
        except IndexError as e:
            raise IndexError(f"idx: {i}, url: {html_url}") from e
        dom_published = dom_episode.find("div", {"class": "studio-episode-published"})
        dom_spans = dom_published.findAll("span")
        views = dom_spans[0].text
        out[id] = views
    return out


def fetch_gabtv_today(channel_name: str, channel_id: str) -> List[VideoInfo]:
    now_datestr = iso_fmt(now_local())
    # GabTV does not PUBLISH VIEWS IN IT'S FEED, so scrape them from the html
    # here.
    out: List[VideoInfo] = []
    view_data = fetch_views(channel_id)
    channel_url = f"https://tv.gab.com/channel/{channel_id}"
    json_feed_url: str = f"https://tv.gab.com/channel/{channel_id}/feed/json"
    # Note: Gab.TV currently returns 403 when using the request lib. The
    # work-around (for now) is to use the curl alternative. Gab should be
    # notified of this limitation and if possible, we should go back to using
    # the request lib to fetch this data.
    json_str = _fetch_html_using_curl(json_feed_url)
    data = json.loads(json_str)
    for item in data["items"]:
        try:
            url: str = item["url"]
            id: str = item["id"]
            title: str = item["title"]
            # Summary is sometimes empty.
            summary: str = item.get("summary", "")
            image: str = item["image"]
            published_date: str = item["date_modified"]
        except KeyError as ke:
            log_error(err_str=f"Error while parsing {channel_id} because of {ke}\n")
            continue

        # GabTV does not PUBLISH VIEWS IN IT'S FEED.
        # print(url)
        views = view_data.get(id)
        if views is None:
            sys.stderr.write(f"{__file__}: Warning, views for {url} is None\n")
            views = "?"
        vi = VideoInfo(
            channel_name=channel_name,
            title=title,
            date_published=published_date,
            date_discovered=now_datestr,
            date_lastupdated=now_datestr,
            channel_url=channel_url,
            source="tv.gab.com",
            url=url,
            duration="?",  # TODO: get this
            description=summary,
            img_src=image,
            iframe_src="",  # No iframe source yet
            views=views,
            profile_img_src="",
        )
        out.append(vi)
    return out


def test() -> None:
    from pprint import pprint  # pylint: disable=import-outside-toplevel

    channel_id: str = input("Enter Channel ID: ")
    channel_name: str = input("Enter Channel Name: ")
    vids: List[VideoInfo] = fetch_gabtv_today(channel_name=channel_name, channel_id=channel_id)
    print("First two videos:")
    vids = vids[0:2]
    for v in vids:
        pprint(v.to_dict())


if __name__ == "__main__":
    test()



"""
    Trampoline to get tests.
"""


def main() -> None:
    """Command hooking point for vidcrawler_tests"""
    import os  # pylint: disable=import-outside-toplevel
    import sys  # pylint: disable=import-outside-toplevel

    os.chdir(os.path.dirname(__file__))
    sys.exit(os.system("pytest --verbose testing"))


if __name__ == "__main__":
    main()



"""
Tests for Playwright.
"""

import os
import unittest

from playwright.sync_api import sync_playwright

IS_GITHUB_RUNNER = os.environ.get("GITHUB_ACTIONS") == "true"


class PlaywrightTester(unittest.TestCase):
    """Tests basic functionality of Playwright."""

    @classmethod
    def setUpClass(cls):
        rtn = os.system("playwright install")
        if rtn != 0:
            raise OSError("Failed to install Playwright.")
        cls.playwright = sync_playwright().start()
        # Launch a headed browser by setting headless to False
        cls.browser = cls.playwright.chromium.launch(headless=IS_GITHUB_RUNNER)

    @classmethod
    def tearDownClass(cls):
        cls.browser.close()
        cls.playwright.stop()

    def setUp(self):
        self.page = self.browser.new_page()

    def tearDown(self):
        self.page.close()

    def test_has_title(self):
        """Simple test to verify the title of a page."""
        # Go to page google.com
        self.page.goto("https://www.google.com")
        # Get page title
        title = self.page.title()
        # Verify the title
        self.assertEqual(title, "Google")


if __name__ == "__main__":
    unittest.main()



# pylint: disable=missing-module-docstring,missing-class-docstring,missing-function-docstring

import unittest

from vidcrawler.spreaker import fetch_spreaker_today


class SpreakerScraperTest(unittest.TestCase):
    def test_fetch_spreaker(self):
        vid_list = fetch_spreaker_today("Sara Carter", "4495281")
        self.assertTrue(vid_list)


if __name__ == "__main__":
    unittest.main()



"""
Scraper for spreaker.com
"""

# pylint: disable=line-too-long,missing-function-docstring,consider-using-f-string,too-many-locals,invalid-name,no-else-return,R0801

import sys
from typing import List

import feedparser  # type: ignore

from vidcrawler.date import iso_fmt, now_local

from .fetch_html import fetch_html_using_request_lib as fetch_html
from .video_info import VideoInfo

# EXPERIMENTAL - parses sara carter from spreaker.


def rss_element_to_video_info(channel_name: str, rss_element: dict) -> VideoInfo:
    thumbnail_img = rss_element["image"]["href"]
    link = rss_element["link"]
    published = rss_element["published"]
    title = rss_element["title"]
    description = rss_element.get("subtitle") or rss_element.get("summary", "")
    # summary = entry['summary']
    duration = rss_element["itunes_duration"]
    now_str = iso_fmt(now_local())
    o = VideoInfo(
        channel_name=channel_name,
        title=title,
        date_published=published,
        date_discovered=now_str,
        date_lastupdated=now_str,
        source="spreaker.com",
        url=link,
        duration=duration,
        description=f"{description}",
        img_src=thumbnail_img,
        iframe_src="",
        views="?",
        profile_img_src=thumbnail_img,
    )
    return o


def fetch_spreaker_today(channel_name: str, channel: str) -> List[VideoInfo]:
    url = f"https://www.spreaker.com/show/{channel}/episodes/feed"
    sys.stdout.write("Spreaker visiting %s (%s)\n" % (channel, url))
    content = fetch_html(url)
    feed = feedparser.parse(content)
    output: List[VideoInfo] = []
    for entry in feed.entries:
        try:
            vid = rss_element_to_video_info(channel_name=channel_name, rss_element=entry)
            output.append(vid)
        except BaseException as err:  # pylint: disable=broad-except
            sys.stderr.write(f"{__file__}: Error while parsing {channel_name} entry:\n {str(entry['title'])}:\n because of {err}")
    return output


if __name__ == "__main__":
    print(fetch_spreaker_today("Sara Carter", "4495281"))
